print("import: 'rope'")
import rope

print("import: 'rope.base'")
import rope.base

print("import: 'rope.base.oi'")
import rope.base.oi

print("import: 'rope.base.oi.type_hinting'")
import rope.base.oi.type_hinting

print("import: 'rope.base.oi.type_hinting.providers'")
import rope.base.oi.type_hinting.providers

print("import: 'rope.base.oi.type_hinting.resolvers'")
import rope.base.oi.type_hinting.resolvers

print("import: 'rope.base.utils'")
import rope.base.utils

print("import: 'rope.contrib'")
import rope.contrib

print("import: 'rope.contrib.autoimport'")
import rope.contrib.autoimport

print("import: 'rope.refactor'")
import rope.refactor

print("import: 'rope.refactor.importutils'")
import rope.refactor.importutils

